﻿/*********************************************************************
 * The MIT License (MIT)
 * 
 * Copyright (c) 2010 Manish Sinha <mail@manishsinha.net>
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:

 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 ********************************************************************/


namespace LpNet.WadlSharp.Common.Custom
{
    using System.CodeDom;
    using System.Collections.Generic;

    class Customizer
    {
        public static CodeTypeDeclaration CreateICustomize()
        {
            CodeTypeDeclaration customizeInterface = new CodeTypeDeclaration("ICustomize");

            customizeInterface.IsInterface = true;
            customizeInterface.Attributes = MemberAttributes.Public | MemberAttributes.Final;

            #region Create Parse Method

            CodeMemberMethod parse = new CodeMemberMethod();

            parse.Name = IdentifierNames.ParseMethodName;
            parse.ReturnType = new CodeTypeReference("T");

            CodeTypeParameter typeParameter = new CodeTypeParameter("T");
            parse.TypeParameters.Add(typeParameter);

            CodeParameterDeclarationExpression param = new CodeParameterDeclarationExpression(typeof(string), "rawString");
            parse.Parameters.Add(param);

            customizeInterface.Members.Add(parse);

            #endregion

            #region Create Fetch Method

            CodeMemberMethod fetch = new CodeMemberMethod();

            fetch.Name = IdentifierNames.FetchMethodName;
            fetch.ReturnType = new CodeTypeReference(typeof(string));

            CodeParameterDeclarationExpression httpMethodParam = new CodeParameterDeclarationExpression(IdentifierNames.HttpMethodTypeEnum, "methodType");
            fetch.Parameters.Add(httpMethodParam);
            CodeParameterDeclarationExpression uriParam = new CodeParameterDeclarationExpression(typeof(string), "uri");
            fetch.Parameters.Add(uriParam);
            CodeParameterDeclarationExpression httpHeadersParam = new CodeParameterDeclarationExpression(typeof(string), "httpHeadersParam");
            fetch.Parameters.Add(httpHeadersParam);
            CodeParameterDeclarationExpression dataParam = new CodeParameterDeclarationExpression(typeof(string), "dataParam");
            fetch.Parameters.Add(dataParam);

            customizeInterface.Members.Add(fetch);

            #endregion

            #region Create Fetch Method

            CodeMemberMethod serialize = new CodeMemberMethod();

            serialize.Name = IdentifierNames.SerializeMethodName;
            serialize.ReturnType = new CodeTypeReference(typeof(string));

            CodeParameterDeclarationExpression serializeData = new CodeParameterDeclarationExpression(IdentifierNames.HttpMethodTypeEnum, "methodType");
            serialize.Parameters.Add(serializeData);
            CodeParameterDeclarationExpression keyvalues = new CodeParameterDeclarationExpression("Dictionary<string, string>", "keyvalues");
            serialize.Parameters.Add(keyvalues);

            customizeInterface.Members.Add(serialize);

            #endregion

            return customizeInterface;
        }

        public static CodeTypeDeclaration CreateHttpMethodTypes()
        {
            string enumName = IdentifierNames.HttpMethodTypeEnum;
            CodeTypeDeclaration httpMethodEnum = new CodeTypeDeclaration(enumName);

            httpMethodEnum.IsEnum = true;

            CodeMemberField getMethod = new CodeMemberField(enumName, "GET");
            httpMethodEnum.Members.Add(getMethod);

            CodeMemberField postMethod = new CodeMemberField(enumName, "POST");
            httpMethodEnum.Members.Add(postMethod);

            CodeMemberField putMethod = new CodeMemberField(enumName, "PUT");
            httpMethodEnum.Members.Add(putMethod);

            CodeMemberField patchMethod = new CodeMemberField(enumName, "PATCH");
            httpMethodEnum.Members.Add(patchMethod);

            return httpMethodEnum;
        }

        public static CodeTypeDeclaration CreateEnumDescription(HashSet<string> namespaces)
        {
            CodeTypeDeclaration enumDesc = new CodeTypeDeclaration(IdentifierNames.EnumDescriptionType);

            enumDesc.Attributes = MemberAttributes.Static | MemberAttributes.Public | MemberAttributes.Final;
            enumDesc.IsClass = true;

            CodeMemberMethod createEnumString = new CodeMemberMethod();

            #region Method Declaration

            createEnumString.Name = IdentifierNames.EnumDescriptionCreateStringName;
            createEnumString.Attributes = MemberAttributes.Public | MemberAttributes.Final;
            createEnumString.ReturnType = new CodeTypeReference(typeof(string));
            createEnumString.TypeParameters.Add(new CodeTypeParameter("T"));

            #endregion

            #region Method Parameters

            CodeParameterDeclarationExpression httpMethodParam = new CodeParameterDeclarationExpression("T", IdentifierNames.EnumDescriptionCreateStringParam);
            createEnumString.Parameters.Add(httpMethodParam);

            enumDesc.Members.Add(createEnumString);

            #endregion

            #region Method Body

            #region Create - FieldInfo fi = currentEnum.GetType().GetField(currentEnum.ToString());

            string fieldInfoFieldName = "fi";
            
            CodeMethodInvokeExpression currentEnumGetType =
                new CodeMethodInvokeExpression(
                    new CodeFieldReferenceExpression(null, IdentifierNames.EnumDescriptionCreateStringParam), "GetType");
            
            CodeMethodInvokeExpression currentEnumToString =
                new CodeMethodInvokeExpression(
                    new CodeFieldReferenceExpression(null, IdentifierNames.EnumDescriptionCreateStringParam), "ToString");

            CodeMethodInvokeExpression getFieldInfo = new CodeMethodInvokeExpression(currentEnumGetType, "GetField",
                                                                                     new CodeExpression[]
                                                                                         {currentEnumToString});

            CodeVariableDeclarationStatement declareFieldInfo = new CodeVariableDeclarationStatement(
                "FieldInfo",
                fieldInfoFieldName,
                getFieldInfo
                );
            createEnumString.Statements.Add(declareFieldInfo);

            #endregion

            #region Create - DescriptionAttribute da = Attribute.GetCustomAttribute(fi, typeof(DescriptionAttribute)) as DescriptionAttribute;

            string descAttributeName = "da";

            CodeMethodInvokeExpression getCustomMethodCall =
                new CodeMethodInvokeExpression(new CodeFieldReferenceExpression(null, "Attribute"), "GetCustomAttribute",
                                               new CodeExpression[]
                                                   {
                                                       new CodeFieldReferenceExpression(null, fieldInfoFieldName),
                                                       new CodeFieldReferenceExpression(null,
                                                                                        "typeof(DescriptionAttribute)")
                                                   });

            CodeCastExpression castToDescAtt = new CodeCastExpression("DescriptionAttribute", getCustomMethodCall);

            CodeVariableDeclarationStatement getDescAtt = new CodeVariableDeclarationStatement(
                "DescriptionAttribute",
                descAttributeName,
                castToDescAtt
                );

            
            createEnumString.Statements.Add(getDescAtt);

            #endregion

            #region Create return (da != null) ? da.Description : currentEnum.ToString();

            string returnSring = string.Format("return ({0} != null) ? {1}.Description : {2}.ToString()",
                                                   descAttributeName, descAttributeName,
                                                   IdentifierNames.EnumDescriptionCreateStringParam);

            CodeSnippetExpression returnStatement = new CodeSnippetExpression(returnSring);

            createEnumString.Statements.Add(returnStatement);

            #endregion

            #endregion

            return enumDesc;
        }
    }
}
